#version 120

//#define Celshade
#define Fog
//#define ShadowColor
#define SSEO
//#define SSAO
//#define WorldTimeAnimation

const int shadowMapResolution = 2048; //[1024 2048 4096]
const float shadowDistance = 256.0;

#define SHADOW_MAP_BIAS 0.9
#define MAX_COLOR_RANGE 32.0
const float sunPathRotation = -40.0;
const float eyeBrightnessHalflife = 10.0f;
const int noiseTextureResolution  = 1024;
const bool shadowHardwareFiltering = true;

varying vec3 lightVector;
varying vec3 upVec;
varying vec3 sunVec;
varying vec3 moonVec;
varying float sunVisibility;
varying float moonVisibility;

varying vec4 texcoord;

uniform sampler2D colortex0;
uniform sampler2D colortex2;
uniform sampler2D colortex4;
uniform sampler2D colortex5;
uniform sampler2D depthtex0;
uniform sampler2D depthtex2;
uniform sampler2D noisetex;

uniform sampler2DShadow shadowtex0;
#ifdef ShadowColor
uniform sampler2DShadow shadowtex1;
uniform sampler2DShadow shadowcolor;
#endif

uniform mat4 gbufferProjection;
uniform mat4 gbufferProjectionInverse;
uniform mat4 gbufferModelViewInverse;
uniform mat4 gbufferModelView;
uniform mat4 shadowProjection;
uniform mat4 shadowModelView;
uniform float rainStrength;
uniform ivec2 eyeBrightnessSmooth;
uniform int isEyeInWater;
uniform int worldTime;
uniform float viewWidth;
uniform float viewHeight;
uniform float aspectRatio;
uniform float far;
uniform float near;
uniform float frameTimeCounter;
uniform vec3 cameraPosition;

float time = float(worldTime);
float tB = max(sin(time/12000*22/7),0.0);
float eBS = eyeBrightnessSmooth.y/240.0;

#ifdef WorldTimeAnimation
float frametime = float(worldTime)/20.0;
#else
float frametime = frameTimeCounter;
#endif

float nolight = 1.0-(clamp((time-12000.0)/500.0,0.0,1.0)-clamp((time-13500.0)/500.0,0.0,1.0) + clamp((time-22500.0)/500.0,0.0,1.0)-clamp((time-23500.0)/500.0,0.0,1.0));

vec3 calcLightColor(vec3 morning, vec3 day, vec3 night, vec3 rain){
vec3 c = mix(night,mix(morning,day,pow(tB,0.75)),sunVisibility);
return mix(c,pow(length(c),0.2)*rain,rainStrength);
}

float getSSS(vec3 normal, vec3 lightVector, vec3 fragpos){
return max(dot(-normal,lightVector),0.0)*pow(max(dot(normalize(fragpos),normalize(lightVector)),0.0),25.0)*(1.0-rainStrength);
}

float ld(float depth) {
   return (2.0 * near) / (far + near - depth * (far - near));
}

float dither4x4(vec2 pos)
{
	const int ditherPattern[16] = int[16](
		0, 8, 2, 10,
		12, 4, 14, 6,
		3, 11, 1, 9,
		15, 7, 13, 5);

    vec2 positon = floor(mod(vec2(texcoord.s * viewWidth,texcoord.t * viewHeight), 4.0f));

	int dither = ditherPattern[int(positon.x) + int(positon.y) * 4];

	return float(dither) / 16.0f;
}

float Blinn_Phong(vec3 ppos, vec3 lvector, vec3 normal, float gloss, float visibility)  {
	vec3 lightDir = vec3(lvector);
	
	vec3 surfaceNormal = normal;
	float cosAngIncidence = dot(surfaceNormal, lightDir);
	cosAngIncidence = clamp(cosAngIncidence, 0.0, 1.0);	
	
	vec3 viewDirection = normalize(-ppos);
	
	vec3 halfAngle = normalize(lightDir + viewDirection);
	float blinnTerm = dot(surfaceNormal, halfAngle);
	
	float normalDotEye = dot(normal, normalize(ppos));
	float fresnel = clamp(pow(1.0 + normalDotEye, 5.0),0.0,1.0);
	fresnel = fresnel*0.85 + 0.15 * (1.0-fresnel);
	float pi = 3.1415927;
	float n =  pow(2.0,gloss);
	return (pow(blinnTerm, n )*((n+8.0)/(8*pi)))*visibility;
}

#include "lib/fragPos.glsl"
#include "lib/shadowPos.glsl"
#include "lib/shadows.glsl"
#include "lib/lightColors.glsl"
#include "lib/torchColor.glsl"
#include "lib/waterColor.glsl"
#include "lib/fogCalc.glsl"
#include "lib/clouds.glsl"
#include "lib/celShading.glsl"
#include "lib/ambientOcclusion.glsl"

void main() {
	
	vec3 color = pow(max(texture2D(colortex0, texcoord.st).rgb,vec3(1.0/255.0)),vec3(2.2));
	vec3 aux = texture2D(colortex4, texcoord.st).rgb;
	float ao = 1.0;
	float spec = 0.0;
	
	float skymap = pow(aux.r,2.2);
	float matflag = aux.g;
	float wmatflag = texture2D(colortex5, texcoord.st).g;
	float torchmap = pow(aux.b,2.2)*0.15 + pow(max(pow(aux.b,2.2)-0.5,0.0)*2.0,5.0);
	
	float land = float(matflag > 0.05);
	float lava = float(matflag > 0.18 && matflag < 0.22);
	float metal = float(matflag > 0.28 && matflag < 0.32);
	float foliage = float(matflag > 0.38 && matflag < 0.42);
	float emissive = float(matflag > 0.48 && matflag < 0.52);
	
	vec3 light = calcLightColor(light_m, light_d + vec3(0.0, 0.25, 1.0)*(1-lava), light_n, light_r);
	vec3 ambient = calcLightColor(sky_m, sky_d, sky_n, sky_r);
	
	float pixeldepth = texture2D(depthtex2,texcoord.xy).x;
	
	vec4 fragpos = getFragPos(texcoord.xy,pixeldepth);
	
	if (land > 0.9) {
		vec3 normal = texture2D(colortex2, texcoord.st).rgb*2.0-1.0;
		//color = vec3(1.0,1.0,1.0)*0.4;
		
		vec4 worldpos = getWorldPos(fragpos);
		float xzDist = worldpos.x * worldpos.x + worldpos.z * worldpos.z;
		float yDist  = worldpos.y * worldpos.y;
		
		vec3 finallight = vec3(0);
		vec3 scenelight = vec3(0);
		vec3 inambient = vec3(0.8)*pow(skymap,0.5)*(0.5+0.5*tB)*(1.0-eBS)*(1.0-rainStrength*0.8)*(1.0-isEyeInWater)*(0.05+0.95*sunVisibility);
		vec3 torchlight = pow(torch_c,vec3(1.2-0.2*pow(aux.b,0.25)))*torchmap*8.0;
		vec3 emissivelight = color*(emissive+lava)*8.0;
		
		float NdotL = max(dot(normal, lightVector)*1.05-0.05,0.0);
		float shading = getShadows(shadowtex0, worldpos, NdotL, foliage);
		shading = max(shading*mix(max(NdotL,0.0), 1.0, foliage)*nolight,0.0);
		
		vec3 shadowcol = vec3(0.0);
		#ifdef ShadowColor
		if (shading < 0.999) shadowcol = getShadowsColor(worldpos, NdotL, foliage)*getShadows(shadowtex1, worldpos, NdotL, foliage)*NdotL;
		#endif
		
		float water = float(wmatflag > 0.08 && wmatflag < 0.12);
		float absorb = float((water > 0.9 && isEyeInWater < 0.9) || (water < 0.9 && isEyeInWater > 0.9));
		vec3 uw_tint = mix(vec3(1.0),pow(water_c,vec3((1.0-skymap)*2.0))*skymap,absorb)*(1.0+2.0*water);
		
		float sss = 0.0;
		if (foliage > 0.9) sss = getSSS(normal,lightVector,fragpos.xyz);
		
		#ifdef SSEO
		if (lava < 0.9 && water < 0.9) ao = sseo(ao,depthtex2,fragpos.xyz);
		#endif
		#ifdef SSAO
		ao = ssao(depthtex2,normal);
		#endif
		
		float minlight = 2.0/255.0 * (4.0-3.0*eBS);
		float lightmult = mix(3.0+2.0*shading,1.0,eBS);
		scenelight = mix(((ambient+inambient)*skymap+minlight)*lightmult*ao,light*lightmult*(1.0+3.0*sss),max(vec3(shading),shadowcol)*(1.0-0.9*rainStrength));
		finallight = scenelight*uw_tint+torchlight*ao+emissivelight;
		
		color *= finallight;
		
		float desat = clamp(max(pow(shading,0.2),skymap)*sunVisibility*(1-rainStrength*0.8)+aux.b+emissive+lava,0.2,1.0);
		vec3 desat_c = mix(vec3(length(light_n))*0.05*sky_r,light_n*2.0,skymap*(1.0-rainStrength));
		color = mix(length(color)*desat_c,color,desat);
		
		#ifdef Fog
		if (isEyeInWater < 0.9 || (isEyeInWater > 0.9 && water > 0.9)) color = calcFog(color, light, fragpos.xyz);
		#endif
		
		spec = Blinn_Phong(fragpos.xyz,lightVector,normal,7+moonVisibility + 4.0*metal,shading*(0.1-0.095*metal));
	}
	else{
	color = mix(length(color)*pow(light_n,vec3(0.8))*2.0,color,sunVisibility);
	color = color*2.0 + getSkyColor(fragpos.xyz, light);
	color = drawCloud(fragpos.xyz,color,light);
	}
	
	#ifdef Celshade
	if (wmatflag > 0.05) color *= celshade(depthtex2,length(fragpos.xyz));
	#endif
	
/* DRAWBUFFERS:0 */
	
	gl_FragData[0] = vec4(pow(color/MAX_COLOR_RANGE,vec3(1.0/2.2)),spec);

}
